import time

from steel_seal import SteelSeal


def test_signature():
    token = SteelSeal.generate_token()
    data = "Hello, World"
    seal_ins = SteelSeal(token)

    sig_info = seal_ins.signature(data)

    print(sig_info)
    print("进行第一次签名验证：{status}".format(status="合法" if seal_ins.verify(data, sig_info) else "非法"))

    time.sleep(4)

    print("模拟重放：{status}".format(status="合法" if seal_ins.verify(data, sig_info) else "非法"))



import random
import string


def get_random_str(num: int) -> str:
    """
    获取指定长度的随机内容
    :param num: 指定获取随机内容的长度
    :return: 返回生成的随机内容
    """
    str_list = [random.choice(string.digits + string.ascii_letters) for i in range(num)]
    random_str = ''.join(str_list)
    return random_str



import time
from pydantic import BaseModel
from typing import List
from .engine import Engine


# 初始化参数
class MemoryOptions(BaseModel):
    # 指定缓存过期时间，单位：秒
    # 默认300秒后历史记录自动过期
    expire_time = 3


# 缓存的数据单元
class CacheUint(BaseModel):
    key: str
    time: int


class MemoryEngine(Engine):
    def __init__(self, *args: MemoryOptions):
        # 用于缓存记录的队列
        self.__list: List[CacheUint] = []
        # 用于查询缓存记录的词典
        self.__dict: dict[str, bool] = {}
        # 最少20秒回收一次过期缓存
        self.__recycling_time = 20
        # 上次执行回收任务时间
        self._last_recycling_time = int(time.time())
        # 初始化配置参数
        # 如果没有指定配置参数则默认记录300秒后自动过期
        self.__options = args[0] if len(args) > 0 else MemoryOptions()

    def add(self, key: str):
        """
        添加记录到缓存列表
        :param key:
        :return:
        """
        if self.__dict.__contains__(key):
            return

        cur_time = int(time.time())

        if cur_time - self._last_recycling_time > self.__recycling_time:
            self.__recycling_cache()

        self.__list.append(CacheUint(
            key=key,
            time=cur_time
        ))
        self.__dict[key] = True

    def exist(self, key: str) -> bool:
        """
        判断记录是否已存在
        :param key:
        :return:
        """
        return self.__dict.__contains__(key)

    def __recycling_cache(self):
        """对过期记录进行释放"""
        cur_time = int(time.time())

        for i in range(len(self.__list)):
            cache = self.__list[i]

            if cache.time + self.__options.expire_time < cur_time and self.__dict.__contains__(cache.key):
                del self.__dict[cache.key]
                continue

            self.__list = self.__list[i:]



from abc import ABC


class Engine(ABC):
    def add(self, key: str):
        """
        添加Key到记录列表
        :param key: 唯一标识符
        :return:
        """

    def exist(self, key: str) -> bool:
        """
        判断当前Key是否已存在
        :param key: 唯一标识符
        :return:
        """


import setuptools


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="steel_seal",
    version="0.0.5",
    author="afeyer",
    author_email="afeyer@h5base.cn",
    description="简单、易用且支持防重放的签名工具",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="",
    packages=setuptools.find_packages(),
    python_requires='>=3.0',  # 对python的最低版本要求
)


import hashlib
import time

from pydantic import BaseModel
from .engine import Engine, MemoryEngine
from .util import get_random_str


class Signature(BaseModel):
    # 当前用于签名的时间戳
    timestamp: int
    # 用于签名的随机内容
    nonce: str
    # 返回签名内容
    signature: str


class SteelSeal:
    """
    SteelSeal Class

    参数：
        token： 必填，用于签名或验签的令牌
        engine： 选填，指定当前实例使用的存储引擎，主要用于防重放。如果不指定，则默认使用MemoryEngine
    """
    def __init__(self, token: str, *args: Engine):
        """初始化签名实例"""
        self.__token = token
        self.__engine = args[0] if len(args) > 0 else MemoryEngine()

    @staticmethod
    def generate_token() -> str:
        """
        生成一个新的Token
        :return: 返回生成的Token
        """
        return get_random_str(32)

    def signature(self, data: str) -> Signature:
        """
        对数据进行签名
        :param data: 需要进行签名的数据
        :return: 签名对象，主要包括timestamp,nonce,signature
        """
        nonce = get_random_str(8)
        timestamp = int(time.time())

        signature = self.__sort_and_hash(self.__token, data, nonce, str(timestamp))
        return Signature(
            timestamp=timestamp,
            nonce=nonce,
            signature=signature
        )

    def verify(self, data: str, sig_info: Signature) -> bool:
        """
        使用token对签名进行验证
        :param data: 用于进行验签的数据，一般是指query或者body参数
        :param sig_info: 从请求的Query中获取的签名对象，主要包括timestamp、nonce、signature
        :return: 签名是否合法
        """
        cur_time = int(time.time())

        # 如果签名时间超过5分钟则判定当前签名已过期
        if sig_info.timestamp and cur_time - sig_info.timestamp > 300:
            return False

        # 如过缓存中存在当前签名记录则判定签名已被使用
        if self.__engine:
            if self.__engine.exist(sig_info.nonce):
                return False
            self.__engine.add(sig_info.nonce)

        return self.__sort_and_hash(self.__token, data, sig_info.nonce, str(sig_info.timestamp)) == sig_info.signature

    def __sort_and_hash(self, *args) -> str:
        """
        对参数进行字典排序并计算Hash
        :param args:
        :return:
        """
        sort_list = list(args)
        sort_list.sort()
        sha = hashlib.sha1()
        sha.update("".join(sort_list).encode())
        return sha.hexdigest()

